/**
 * 
 */
package it.unicam.cs.asdl1617.queens;

import java.util.ArrayList;
import java.util.List;

/**
 * Risolutore del problema delle 8 regine.
 * 
 * @author luca
 *
 */
public class EightQueensSolver {

    // Scacchiera privata di questo solver
    private Scacchiera s;

    // Contiene l'ultima soluzione trovata;
    // quando è null allora sono all'inizio della ricerca
    private List<PositionedQueen> lastSolution;

    // flag per indicare quando ho esplorato tutto lo spazio delle soluzioni
    private boolean terminated;

    // Registra lo stato corrente durante la ricerca delle soluzioni
    // Uno stato consiste di un certo numero di regine (almeno una) posizionate
    // in modo che non si mangino una con l'altra.
    // Se lo stato contiene otto regine posizionate allora è una soluzione.
    private List<PositionedQueen> currentState;

    /**
     * Crea un nuovo solver
     */
    public EightQueensSolver() {
        s = new Scacchiera();
        this.lastSolution = null;
        this.terminated = false;
    }

    /**
     * Restituisce la soluzione successiva nella ricerca con backtracking.
     * 
     * @return la prossima soluzione oppure null se non ci sono più soluzioni.
     */
    public List<PositionedQueen> nextSolution() {
        // Ritorno sempre null dopo che sono terminato
        if (this.terminated)
            return null;
        // se non ho terminato la ricerca
        this.privateNext();
        // il metodo recNext mette a true il flag terminated
        // quando non ci sono più soluzioni
        if (this.terminated)
            return null;
        else
            return this.lastSolution;
    }

    /*
     * Se this.lastSolution è null allora siamo all'inizio e quindi viene
     * iniziata la ricerca a partire dalla casella (0,0). Altrimenti, se non è
     * null, il metodo cerca la prossima soluzione a partire dall'ultima
     * soluzione trovata.
     */
    private void privateNext() {
        if (this.lastSolution == null) {
            // Siamo all'inizio
            // Posizioniamo la prima regina in (0,0)
            this.s.putQueen(0, 0);
            this.currentState = new ArrayList<PositionedQueen>();
            this.currentState.add(new PositionedQueen(0, 0));
        }
        // this.currentState contiene lo stato corrente
        do {
            // faccio almeno un altro passo
            this.nextStep();
            // Controllo se sono terminato
            if (this.terminated)
                return; // interrompo la ricerca
        } while (this.currentState.size() < 8);
        // this.currentState contiene una soluzione
        // Copio this.currentState dentro this.lastSolution
        this.lastSolution = new ArrayList<PositionedQueen>();
        for (PositionedQueen p : this.currentState)
            this.lastSolution.add(new PositionedQueen(p.getI(), p.getJ()));
    }

    /*
     * Dato uno stato corrente, o decide di terminare la ricerca oppure
     * determina il prossimo stato da esplorare. Il prossimo stato da esplorare
     * potrà essere o uno stato in cui ho posizionato una nuova regina oppure
     * uno stato in cui ho fatto un backtracking e quindi ho tolto alcune regine
     * posizionate in precedenza e ho riposizionato l'ultima regina tolta nella
     * stessa riga.
     */
    private void nextStep() {
        int r, c; // variabili di appoggio per riga e colonna
        // provo ad andare avanti, cioè a posizionare la prossima regina
        if (this.currentState.size() < 8) {
            // provo a posizionare una regina nella riga
            // data da this.currentState.size()
            r = this.currentState.size();
            c = this.s.putInLineAfterColumn(r, 0);
            if (c >= 0) {
                // posso mettere la regina e ho finito
                this.currentState.add(new PositionedQueen(r, c));
                this.s.putQueen(r, c);
                return;
            }
            // Non posso posizionare una nuova regina, 
            // quindi faccio backtracking
        }
        /*
         * Backtracking: Ciclo togliendo l'ultima regina nello stato corrente e
         * provando a riposizionarla sulla stessa riga
         */
        do {
            // riga dell'ultima regina posizionata
            r = this.currentState.get(this.currentState.size() - 1).getI();
            // colonna dell'ultima regina posizionata
            c = this.currentState.get(this.currentState.size() - 1).getJ();
            // tolgo l'ultima regina posizionata
            this.currentState.remove(this.currentState.size() - 1);
            this.s.removeQueen(r, c);
            // provo a riposizionarla nella stessa riga r
            int newC = this.s.putInLineAfterColumn(r, c + 1);
            if (newC >= 0) {
                // Posso riposizionare la regina e ho terminato
                this.s.putQueen(r, newC);
                this.currentState.add(new PositionedQueen(r, newC));
                return;
            }
            // Non sono riuscito a riposizionare la regina nella stessa riga
            // Riciclo sulla regina precedenta a meno che non sia r == 0
        } while (r > 0); 
        // E' stata tolta la regina in riga 0 e non sono riuscito a
        // riposizionarla in riga zero.
        // Quindi ho terminato la ricerca
        this.terminated = true;
    }
}
